import asyncio
import os
from types import ModuleType
from typing import Any, Optional
from sqlalchemy import Engine
from sqlmodel import create_engine as alchemy_create_engine
from sqlalchemy.ext.asyncio import (
    create_async_engine as alchemy_create_engine_async,
    AsyncEngine,
    AsyncSession,
)
from sqlalchemy.orm import Session
from contextvars import ContextVar

engines_sync: ContextVar[dict[str, Engine]] = ContextVar("engines_sync")
engines_async: ContextVar[dict[str, AsyncEngine]] = ContextVar("engines_async")


def create_engine_sync(
    module,
    username: str = "",
    password: str = "",
    host: str = "",
    port: str = "",
    **kwargs,
):
    """
    Create a synchronous SQLAlchemy engine based on the provided module and connection parameters.

    Args:
        module: The module object representing the database driver.
        username: The username for the database connection (default: "").
        password: The password for the database connection (default: "").
        host: The host address for the database connection (default: "").
        port: The port number for the database connection (default: "").
        **kwargs: Additional keyword arguments to be passed to the SQLAlchemy create_engine function.

    Returns:
        An SQLAlchemy engine object.

    Raises:
        None.
    """
    # get the name
    db_name = module.__name__.split(".")[-1].lower()
    dialect = module.__name__.split(".")[-2].lower()
    db_path = (
        os.path.dirname(module.__file__).replace(os.getcwd(), "").replace("\\", "/")
    )
    # connect to database

    if dialect == "sqlite":
        return alchemy_create_engine(f"{dialect}://{db_path}/{db_name}.db", **kwargs)
    else:
        return alchemy_create_engine(
            f"{dialect}://{username}:{password}@{host}:{port}/{db_name}", **kwargs
        )


def create_engine_async(module, driver, username: str = "", password: str = "", host: str = "", port: str = "", **kwargs) -> AsyncEngine:  # type: ignore
    """
    Create an asynchronous SQLAlchemy engine based on the provided module, driver, and connection parameters.

    Args:
        module: The module object representing the database dialect.
        driver: The database driver to be used.
        username: The username for the database connection (optional).
        password: The password for the database connection (optional).
        host: The host address for the database connection (optional).
        port: The port number for the database connection (optional).
        **kwargs: Additional keyword arguments to be passed to the SQLAlchemy engine.

    Returns:
        An asynchronous SQLAlchemy engine object.

    Raises:
        None.

    """
    db_name = module.__name__.split(".")[-1].lower()
    dialect = module.__name__.split(".")[-2].lower()
    db_path = (
        os.path.dirname(module.__file__).replace(os.getcwd(), "").replace("\\", "/")
    )
    if dialect == "sqlite":
        return alchemy_create_engine_async(
            f"{dialect}+{driver}://{db_path}/{db_name}.db", **kwargs
        )
    else:
        return alchemy_create_engine_async(
            f"{dialect}+{driver}://{username}:{password}@{host}:{port}/{db_name}",
            **kwargs,
        )


def init_database_sync(engine, module):
    """
    Initializes the database by creating all the tables defined in the SQLModel metadata.

    Args:
        engine (sqlalchemy.engine.Engine): The SQLAlchemy engine object.
        module (module): The module containing the SQLModel metadata.

    Returns:
        None
    """
    module.SQLModel.metadata.create_all(engine)


async def init_database_async(engine: AsyncEngine, module: Any) -> None:
    """
    Initialize the database asynchronously.

    Args:
        engine (AsyncEngine): The SQLAlchemy async engine instance.
        module (Any): The module containing the SQLModel metadata.

    Returns:
        None
    """
    async with engine.begin() as connection:
        await connection.run_sync(module.SQLModel.metadata.create_all)


def createSyncEngine(module: ModuleType, **kwargs):
    """
    Create a new engine and add it to the context variable.

    Args:
        module (ModuleType): The module to create the engine for.
        **kwargs: Additional keyword arguments to pass to the create_engine_sync function.

    Returns:
        None
    """

    engine = create_engine_sync(module, **kwargs)

    engines_dict: dict[str, Engine] = engines_sync.get({})

    engines_dict[module.__name__.split(".")[-1].lower()] = engine

    engines_sync.set(engines_dict)


def createAsyncEngine(module: ModuleType, driver: str, **kwargs):
    """
    Create a new engine and add it to the context variable.

    Args:
        module (ModuleType): The module containing the database driver.
        driver (str): The name of the database driver.
        **kwargs: Additional keyword arguments to be passed to the create_engine_async function.

    Returns:
        None
    """
    engine = create_engine_async(module, driver, **kwargs)

    engines_dict: dict[str, AsyncEngine] = engines_async.get({})

    engines_dict[module.__name__.split(".")[-1].lower()] = engine

    engines_async.set(engines_dict)
    return


def getSyncEngine(database: ModuleType | str) -> Engine:
    """
    Retrieves the sync engine for the specified database.

    Parameters:
        database (ModuleType | str): The database module or name.

    Returns:
        Engine: The sync engine for the specified database.
    """
    if isinstance(database, str):
        engine: Engine = engines_sync.get()[database.lower()]
    elif isinstance(database, ModuleType):
        engine: Engine = engines_sync.get()[database.__name__.split(".")[-1].lower()]
    return engine


def getAsyncEngine(database: ModuleType | str) -> AsyncEngine:
    """
    Retrieves the asynchronous engine for the specified database.

    Args:
        database (ModuleType | str): The database module or name.

    Returns:
        AsyncEngine: The asynchronous engine for the specified database.
    """
    if isinstance(database, str):
        engine: AsyncEngine = engines_async.get()[database.lower()]
    elif isinstance(database, ModuleType):
        engine: AsyncEngine = engines_async.get()[
            database.__name__.split(".")[-1].lower()
        ]
    return engine


async def asyncDbSession(database: ModuleType | str) -> AsyncSession:
    """
    Create an asynchronous database session.

    Args:
        database (ModuleType | str): The database module or the name of the database.

    Returns:
        AsyncSession: The asynchronous database session.

    """
    # get the engine
    if isinstance(database, str):
        engine: AsyncEngine = getAsyncEngine(database)
    elif isinstance(database, ModuleType):
        engine: AsyncEngine = getAsyncEngine(database.__name__.split(".")[-1].lower())
    return AsyncSession(engine)


def syncDbSession(database: ModuleType | str) -> Session:
    """
    Create and return a synchronized database session.

    Parameters:
    - database: Either a string representing the name of the database or a ModuleType object.
                If a string is provided, it will be used to retrieve the engine.
                If a ModuleType object is provided, its name will be used to retrieve the engine.

    Returns:
    - Session: A synchronized database session.

    """
    # get the engine
    if isinstance(database, str):
        engine: Engine = getSyncEngine(database)
    elif isinstance(database, ModuleType):
        engine: Engine = getSyncEngine(database.__name__.split(".")[-1].lower())

    return Session(engine)



from math import e
import os
from threading import Thread
from types import ModuleType
from typing import Any, Generator, Union
import aiofiles
from starlette.middleware.base import BaseHTTPMiddleware
from fastapi import HTTPException, Request, Response
from jinja2 import Environment, FileSystemLoader, Template
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi import HTTPException
from starlette.exceptions import HTTPException
import importlib.util
import asyncio

VANGUARD_CONVENTIONS: list[str] = [
    "layout.html",
    "page.html",
    "script.py",
    "load.py",
    "404.html",
]

VanguardRoutes = dict[str, dict[str, Union[ModuleType, Template, str, Any]]]


class VanguardMiddleware(BaseHTTPMiddleware):
    base_template: Template
    error_template: Template

    def __init__(
        self,
        app,
        route_dir: str = "routes",
        static_dir: str = "static",
        pyodide_dir: str = "pyodide",
        enable_pyodide: bool = True,
        base_html: str = "index.html",
        error_html: str = "error.html",
    ) -> None:
        super().__init__(app)

        self.routes: VanguardRoutes = {}

        async def _init_async():
            self.base_template = await self.loadTemplate(base_html, static_dir)
            self.error_template = await self.loadTemplate(error_html, static_dir)
            for root, convention, route in self.walk(route_dir):
                if convention in [x for x in VANGUARD_CONVENTIONS if ".py" in x]:
                    module: ModuleType | str = await self.loadPython(convention, root)
                    if module:
                        self.routes.setdefault(route, {}).update({convention: module})
                if convention in [x for x in VANGUARD_CONVENTIONS if ".html" in x]:
                    template: Template = await self.loadTemplate(convention, root)
                    if template:
                        self.routes.setdefault(route, {}).update({convention: template})

        # run the async function in a separate thread
        try:
            _init_thread = Thread(target=asyncio.run, args=(_init_async(),))
            _init_thread.start()
            _init_thread.join()
        except Exception as e:
            raise e

    async def loadPython(self, name: str, directory: str) -> ModuleType | str:
        if not os.path.exists(directory + f"/{name}"):
            raise OSError(
                f"Python file '{name}' does not exist in directory '{directory}'"
            )

        if name == "script.py":
            async with aiofiles.open(directory + f"/{name}", mode="r") as f:
                return await f.read()

        spec = importlib.util.spec_from_file_location(name, directory + f"/{name}")
        module = importlib.util.module_from_spec(spec)  # type: ignore
        try:
            spec.loader.exec_module(module)  # type: ignore
        except Exception as e:
            raise OSError(
                f"Error loading module '{name}' in directory '{directory}'"
            ) from e
        return module

    async def loadTemplate(self, name: str, directory: str) -> Template:
        if not os.path.exists(directory + f"/{name}"):
            raise OSError(
                f"Template file '{name}' does not exist in directory '{directory}'"
            )
        template = Environment(
            loader=FileSystemLoader(directory),
        ).get_template(name)
        return template

    def walk(self, directory: str) -> Generator[tuple[str, str, str], Any, None]:
        for root, _, files in os.walk(directory):
            for convention in VANGUARD_CONVENTIONS:
                if convention in files:
                    root = root.rstrip("/") + "/"
                    if root == "/":
                        yield root, convention, root
                    else:
                        yield root, convention, root.replace(directory, "").replace(
                            "\\", "/"
                        ).replace("[", "{").replace("]", "}").replace("\\routes\\", "/")

    async def dispatch(
        self, request: Request, call_next
    ) -> HTMLResponse | JSONResponse | Response | None:
        try:
            url_path = (
                request.url.path + "/"
                if not request.url.path.endswith("/")
                else request.url.path
            )
            if url_path in self.routes:
                if (
                    request.headers.get("X-Requested-With") == "XMLHttpRequest"
                    and request.headers.get("Content-Type") == "application/python"
                ):
                    return await self.renderHTML(request, "partial")
                else:
                    if "text/html" in request.headers.get("accept", "").split(","):
                        return await self.renderHTML(request, "full")
                if url_path not in self.routes:
                    raise HTTPException(detail="Page Not Found", status_code=404)
            else:
                return await call_next(request)
        except HTTPException as e:
            return HTMLResponse(
                content=self.base_template.render(
                    request=request, error_message=e.detail, error_code=e.status_code
                ),
                status_code=e.status_code,
            )

    async def render(
        self, url_path: str, request: Request, mode: str = "full"
    ) -> Union[HTMLResponse, JSONResponse]:
        loaded_data: dict = {}

        if "load.py" in self.routes[url_path]:
            loaded_data = await self.routes[url_path]["load.py"].load(  # type: ignore
                request
            )
        html = (
            self.routes[url_path]["page.html"].render(**loaded_data)  # type: ignore
            if loaded_data
            else self.routes[url_path]["page.html"].render()  # type: ignore
        )

        components = url_path.strip("/").split("/")
        routes = [
            "/" + "/".join(components[: i + 1]) + "/"
            for i in range(len(components))
            if i != 0
        ][::-1]
        routes = [route for route in routes if route != "//"]

        for route in routes:
            if "layout.html" in self.routes[route]:
                html = self.routes[route]["layout.html"].render(slot=html)  # type: ignore

        if url_path == "/" and "layout.html" in self.routes[url_path]:
            html = self.routes[url_path]["layout.html"].render(slot=html)  # type: ignore
        if "script.py" in self.routes[url_path]:
            if mode == "full":
                return HTMLResponse(
                    content=self.base_template.render(
                        slot=html,
                        script=self.routes[url_path]["script.py"] or "",
                        **loaded_data,
                    ),
                    status_code=200,
                )
            else:
                return JSONResponse(
                    content={
                        "body": html,
                        "script": self.routes[url_path]["script.py"] or "",
                        **loaded_data,
                    }
                )
        else:
            if mode == "full":
                return HTMLResponse(
                    content=self.base_template.render(
                        slot=html,
                        **loaded_data,
                    ),
                    status_code=200,
                )
            else:
                return JSONResponse(
                    content={
                        "body": html,
                        **loaded_data,
                    }
                )

    async def renderHTML(
        self, request: Request, mode: str
    ) -> HTMLResponse | JSONResponse:
        url_path = (
            request.url.path + "/"
            if request.url.path != "/" and request.url.path[-1] != "/"
            else request.url.path
        )
        try:
            if "page.html" not in self.routes[url_path]:
                raise HTTPException(status_code=404)
            return await self.render(url_path, request, mode)
        except HTTPException as e:
            error_template = self.routes[url_path].get("404.html")
            if not error_template:
                error_template = self.error_template
            return HTMLResponse(
                content=error_template.render(request=request, error_message=e.detail),  # type: ignore
                status_code=404,
            )
